import json

import pytest
import numpy as np
import pandas as pd
import scipy.spatial.distance as scipy_distance

from whatlies import Embedding, EmbeddingSet
from .common import validate_plot_general_properties


"""
*Guide*

Here are the plot's propertites which could be checked (some of them may not be applicable
for a particular plot/test case):
    - type: the type of plot; usually it's scatter plot with circle marks.
    - data_field: the name of the field of chart data which is used for datapoints' coordinates.
    - data: the position (i.e. coordinates) of datapoints in the plot.
    - x_label: label of x-axis.
    - y_label: label of y-axis.
    - tilte: title of the plot.
    - label_field: the name of the field of chart data which is used for annotating data points with text labels.
    - label: the text labels used for annotation of datapoints.
    - color_field: the name of the field of chart data which is used for coloring datapoints.
"""


@pytest.fixture
def embset():
    names = ["red", "blue", "green", "yellow", "white"]
    vectors = np.random.rand(5, 4) * 10 - 5
    embeddings = [Embedding(name, vector) for name, vector in zip(names, vectors)]
    return EmbeddingSet(*embeddings)


def test_default(embset):
    p = embset.plot_interactive()
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": embset.to_X()[:, :2],
        "x_label": "Dimension 0",
        "y_label": "Dimension 1",
        "title": "Dimension 0 vs. Dimension 1",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)

    # Check if it's an interactive plot (done only in this test)
    assert "selection" in chart["layer"][0]
    # Check tooltip data (only done in this test case)
    tooltip_fields = set(
        [
            chart["layer"][0]["encoding"]["tooltip"][0]["field"],
            chart["layer"][0]["encoding"]["tooltip"][1]["field"],
        ]
    )
    assert tooltip_fields == set(["name", "original"])


def test_int_axis(embset):
    p = embset.plot_interactive(x_axis=2, y_axis=0, x_label="xaxis", title="some chart")
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": np.concatenate([embset.to_X()[:, 2:3], embset.to_X()[:, :1]], axis=-1),
        "x_label": "xaxis",
        "y_label": "Dimension 0",
        "title": "some chart",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_int_axis_with_common_str_axis_metric(embset):
    p = embset.plot_interactive(x_axis=1, y_axis=2, axis_metric="cosine_similarity")
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": embset.to_X()[:, 1:3],
        "x_label": "Dimension 1",
        "y_label": "Dimension 2",
        "title": "Dimension 1 vs. Dimension 2",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_str_axis(embset):
    p = embset.plot_interactive(x_axis="red", y_axis="blue")
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e > embset["red"], e > embset["blue"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "red",
        "y_label": "blue",
        "title": "red vs. blue",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_str_axis_with_common_str_axis_metric(embset):
    p = embset.plot_interactive(
        x_axis="red",
        y_axis="blue",
        y_label="blue_cosine",
        axis_metric="cosine_distance",
        color="name",
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                scipy_distance.cosine(e.vector, embset["red"].vector),
                scipy_distance.cosine(e.vector, embset["blue"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "red",
        "y_label": "blue_cosine",
        "title": "red vs. blue",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "name",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_str_axis_with_different_axis_metric(embset):
    p = embset.plot_interactive(
        x_axis="red", y_axis="blue", axis_metric=[np.dot, "euclidean"]
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                np.dot(e.vector, embset["red"].vector),
                scipy_distance.euclidean(e.vector, embset["blue"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "red",
        "y_label": "blue",
        "title": "red vs. blue",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_emb_axis(embset):
    p = embset.plot_interactive(x_axis=embset["yellow"], y_axis=embset["white"])
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e > embset["yellow"], e > embset["white"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "yellow",
        "y_label": "white",
        "title": "yellow vs. white",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_emb_axis_with_common_str_axis_metric(embset):
    p = embset.plot_interactive(
        x_axis=embset["red"],
        y_axis=embset["green"],
        axis_metric="cosine_similarity",
        annot=False,
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                1 - scipy_distance.cosine(e.vector, embset["red"].vector),
                1 - scipy_distance.cosine(e.vector, embset["green"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "red",
        "y_label": "green",
        "title": "red vs. green",
        "color_field": "",
        # Not applicable: label_field, label
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [chart["encoding"]["x"]["field"], chart["encoding"]["y"]["field"]] == props[
        "data_field"
    ]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["encoding"]["color"]["field"] == props["color_field"]
    assert "text" not in chart["encoding"]
    assert "layer" not in chart
    validate_plot_general_properties(chart, props)


def test_emb_axis_with_different_axis_metric(embset):
    p = embset.plot_interactive(
        x_axis=embset["blue"], y_axis=embset["yellow"], axis_metric=[None, "euclidean"]
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                e > embset["blue"],
                scipy_distance.euclidean(e.vector, embset["yellow"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "blue",
        "y_label": "yellow",
        "title": "blue vs. yellow",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)


def test_mixed_axis(embset):
    p = embset.plot_interactive(x_axis=3, y_axis="white")
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e.vector[3], e > embset["white"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data_field": ["x_axis", "y_axis"],
        "data": vectors,
        "x_label": "Dimension 3",
        "y_label": "white",
        "title": "Dimension 3 vs. white",
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "color_field": "",
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["data"]["name"]])
    assert [
        chart["layer"][0]["encoding"]["x"]["field"],
        chart["layer"][0]["encoding"]["y"]["field"],
    ] == props["data_field"]
    assert np.array_equal(chart_data[["x_axis", "y_axis"]].values, props["data"])
    assert chart["layer"][0]["encoding"]["color"]["field"] == props["color_field"]
    assert chart["layer"][1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data["original"].values, props["label"])
    validate_plot_general_properties(chart["layer"][0], props)



from operator import add, rshift, sub, or_

import pytest
import numpy as np
from spacy.vocab import Vocab
from spacy.language import Language

from whatlies import Embedding, EmbeddingSet
from whatlies.language import SpacyLanguage


@pytest.fixture()
def lang():
    vector_data = {
        k: np.random.normal(0, 1, (2,))
        for k in ["red", "blue", "cat", "dog", "green", "purple"]
    }
    vector_data["cat"] += 21.0
    vector_data["dog"] += 20.0
    vocab = Vocab(strings=vector_data.keys())
    for word, vector in vector_data.items():
        vocab.set_vector(word, vector)
    nlp = Language(vocab=vocab)
    return SpacyLanguage(nlp)


def test_embeddingset_creation():
    foo = Embedding("foo", [0, 1])
    bar = Embedding("bar", [1, 1])

    emb = EmbeddingSet(foo)
    assert len(emb) == 1
    assert "foo" in emb
    emb = EmbeddingSet(foo, bar)
    assert len(emb) == 2
    assert "foo" in emb
    assert "bar" in emb
    emb = EmbeddingSet({"foo": foo})
    assert len(emb) == 1
    assert "foo" in emb


def test_embset_creation_error():
    foo = Embedding("foo", [0, 1])
    # This vector has a different dimension. No bueno.
    bar = Embedding("bar", [1, 1, 2])
    with pytest.raises(ValueError):
        EmbeddingSet(foo, bar)


def test_embset_creation_warning():
    foo = Embedding("foo", [0, 1])
    # This vector has the same name dimension. Dangerzone.
    bar = Embedding("foo", [1, 2])
    with pytest.raises(Warning):
        EmbeddingSet(foo, bar)


@pytest.mark.parametrize("operator", [add, rshift, sub, or_])
def test_artificial_embset(lang, operator):
    emb = lang[["red", "blue", "orange"]]
    v1 = operator(emb["red"], emb["blue"])
    v2 = operator(lang["red"], lang["blue"])
    assert np.array_equal(v1.vector, v2.vector)


def test_merge_basic(lang):
    emb1 = lang[["red", "blue", "orange"]]
    emb2 = lang[["pink", "purple", "brown"]]
    assert len(emb1.merge(emb2)) == 6


def test_average(lang):
    emb = lang[["red", "blue", "orange"]]
    av = emb.average()
    assert av.name == "EmbSet.average()"
    v1 = av.vector
    v2 = (lang["red"].vector + lang["blue"].vector + lang["orange"].vector) / 3
    assert np.array_equal(v1, v2)


def test_to_x_y():
    foo = Embedding("foo", [0.1, 0.3])
    bar = Embedding("bar", [0.7, 0.2])
    buz = Embedding("buz", [0.1, 0.9])
    bla = Embedding("bla", [0.2, 0.8])

    emb1 = EmbeddingSet(foo, bar).add_property("label", lambda d: "group-one")
    emb2 = EmbeddingSet(buz, bla).add_property("label", lambda d: "group-two")
    emb = emb1.merge(emb2)

    X, y = emb.to_X_y(y_label="label")
    assert X.shape == emb.to_X().shape == (4, 2)
    assert list(y) == ["group-one", "group-one", "group-two", "group-two"]


def test_embset_similar_simple_len(lang):
    emb = lang[["red", "blue", "orange"]]
    assert len(emb.embset_similar("red", 1)) == 1
    assert len(emb.embset_similar("red", 2)) == 2


def test_embset_similar_simple_contains(lang):
    emb = lang[["red", "blue", "orange", "cat", "dog"]]
    subset_cat = emb.embset_similar("cat", 2, metric="euclidean").embeddings.keys()
    assert "cat" in subset_cat
    assert "dog" in subset_cat


def test_embset_similar_simple_distance(lang):
    emb = lang[["red", "blue", "orange", "cat", "dog"]]
    emb_red, score_red = emb.score_similar("red", 5)[0]
    assert np.isclose(score_red, 0.0, atol=0.0001)


def test_embset_raise_value_error_n(lang):
    emb = lang[["red", "blue", "orange", "cat", "dog"]]
    with pytest.raises(ValueError):
        emb.score_similar("red", 10)


def test_embset_raise_value_error_emb(lang):
    emb = lang[["red", "blue", "orange", "cat", "dog"]]
    with pytest.raises(ValueError):
        emb.score_similar("dinosaurhead", 1)


def test_corrplot_raise_error(lang):
    with pytest.raises(ValueError):
        emb = lang[["red", "blue", "orange", "pink", "purple", "brown"]]
        emb.plot_correlation(metric="dinosaurhead")


def test_filter(lang):
    emb = lang[["red", "blue", "orange", "pink", "purple", "brown"]]
    assert len(emb) == 6
    assert len(emb.filter(lambda e: "pink" not in e.name)) == 5
    assert len(emb.filter(lambda e: "pink" in e.name)) == 1


def test_pipe(lang):
    embset = lang[["red", "blue", "orange", "pink", "purple", "brown"]]
    assert embset.pipe(len) == 6


def test_to_names_X(lang):
    words = ["red", "blue", "dog"]
    embset = lang[words]
    names, X = embset.to_names_X()
    assert names == words
    assert np.array_equal(X, embset.to_X())


def test_from_names_X():
    names = ["foo", "bar", "buz"]
    X = [
        [1.0, 2],
        [3, 4.0],
        [0.5, 0.6],
    ]
    embset = EmbeddingSet.from_names_X(names, X)
    assert "foo" in embset
    assert len(embset) == 3
    assert np.array_equal(embset.to_X(), np.array(X))

    names = names[:2]
    with pytest.raises(ValueError, match="The number of given names"):
        EmbeddingSet.from_names_X(names, X)


def test_ndim(lang):
    embset = lang[["red", "blue", "dog"]]
    assert embset.ndim == 2


def test_compare_against(lang):
    embset = lang[["red", "blue", "cat"]]
    compared = embset.compare_against(lang["green"])
    true_values = np.array(
        [
            embset["red"] > lang["green"],
            embset["blue"] > lang["green"],
            embset["cat"] > lang["green"],
        ]
    )
    assert np.array_equal(compared, true_values)

    # Test with custom mapping function
    compared = embset.compare_against("cat", mapping=np.dot)
    true_values = np.array(
        [
            np.dot(embset["red"].vector, lang["cat"].vector),
            np.dot(embset["blue"].vector, lang["cat"].vector),
            np.dot(embset["cat"].vector, lang["cat"].vector),
        ]
    )
    assert np.array_equal(compared, true_values)

    # Test with non-existent name or invalid mapping
    with pytest.raises(KeyError):
        embset.compare_against("purple")
    with pytest.raises(ValueError, match="Unrecognized mapping value/type."):
        embset.compare_against(lang["green"], mapping="cosine")


def test_add_property():
    foo = Embedding("foo", [0.1, 0.3, 0.10])
    bar = Embedding("bar", [0.7, 0.2, 0.11])
    emb = EmbeddingSet(foo, bar)
    emb_with_property = emb.add_property("prop_a", lambda d: "prop-one")
    assert all([e.prop_a == "prop-one" for e in emb_with_property])


def test_assign():
    foo = Embedding("foo", [0.1, 0.3, 0.10])
    bar = Embedding("bar", [0.7, 0.2, 0.11])
    emb = EmbeddingSet(foo, bar)
    emb_with_property = emb.assign(
        prop_a=lambda d: "prop-one", prop_b=lambda d: "prop-two"
    )
    assert all([e.prop_a == "prop-one" for e in emb_with_property])
    assert all([e.prop_b == "prop-two" for e in emb_with_property])



import pytest

from whatlies.error import NotInstalled


def test_import_error():
    dinosaurhead = NotInstalled("DinosaurHead", "dinosaur")
    with pytest.raises(ModuleNotFoundError) as e:
        dinosaurhead()
    assert "pip install whatlies[dinosaur]" in str(e.value)



import warnings

from umap import UMAP
from numba import NumbaPerformanceWarning

from whatlies.transformers import Transformer
from whatlies import EmbeddingSet
from whatlies.transformers._common import new_embedding_dict


class Umap(Transformer):
    """
    This transformer transformers all vectors in an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]
    by means of umap. We're using the implementation in [umap-learn](https://umap-learn.readthedocs.io/en/latest/).

    Arguments:
        n_components: the number of compoments to create/add
        kwargs: keyword arguments passed to the UMAP algorithm

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Umap

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "blue", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Umap(3)).plot_interactive_matrix(0, 1, 2)
    ```
    """

    def __init__(self, n_components=2, **kwargs):
        super().__init__()
        self.n_components = n_components
        self.kwargs = kwargs
        self.tfm = UMAP(n_components=n_components, **kwargs)

    def fit(self, embset):
        names, X = embset.to_names_X()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
            self.tfm.fit(X)
        self.is_fitted = True
        return self

    def transform(self, embset):
        names, X = embset.to_names_X()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
            new_vecs = self.tfm.transform(X)
        new_dict = new_embedding_dict(names, new_vecs, embset)
        return EmbeddingSet(new_dict, name=f"{embset.name}.umap_{self.n_components}()")



from typing import Union, Optional, Sequence, Callable
from copy import deepcopy

import numpy as np
import scipy.spatial.distance as scipy_distance
from sklearn.metrics import pairwise_distances

from whatlies.common import handle_2d_plot


class Embedding:
    """
    This object represents a word embedding. It contains a vector and a name.

    Arguments:
        name: the name of this embedding, includes operations
        vector: the numerical representation of the embedding
        orig: original name of embedding, is left alone

    Usage:

    ```python
    from whatlies.embedding import Embedding

    foo = Embedding("foo", [0.1, 0.3])
    bar = Embedding("bar", [0.7, 0.2])

    foo | bar
    foo - bar + bar
    ```
    """

    def __init__(self, name, vector, orig=None):
        self.orig = name if not orig else orig
        self.name = name
        self.vector = np.array(vector)

    def add_property(self, name, func):
        result = self.copy()
        setattr(result, name, func(result))
        return result

    @property
    def ndim(self):
        """
        Return the dimension of embedding vector.
        """
        return self.vector.shape[0]

    def copy(self):
        """
        Returns a deepcopy of the embdding.
        """
        return deepcopy(self)

    def __add__(self, other) -> "Embedding":
        """
        Add two embeddings together.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo + bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} + {other.name})"
        copied.vector = self.vector + other.vector
        return copied

    def __sub__(self, other):
        """
        Subtract two embeddings.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo - bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} - {other.name})"
        copied.vector = self.vector - other.vector
        return copied

    def __neg__(self):
        """
        Negate an embedding.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])

        assert (- foo).vector == - foo.vector
        ```
        """
        copied = deepcopy(self)
        copied.name = f"(-{self.name})"
        copied.vector = -self.vector
        return copied

    def __gt__(self, other):
        """
        Measures the size of one embedding to another one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo > bar
        ```
        """
        return (self.vector.dot(other.vector)) / (other.vector.dot(other.vector))

    def __rshift__(self, other):
        """
        Maps an embedding unto another one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo >> bar
        ```
        """
        copied = deepcopy(self)
        new_vec = (
            (self.vector.dot(other.vector))
            / (other.vector.dot(other.vector))
            * other.vector
        )
        copied.name = f"({self.name} >> {other.name})"
        copied.vector = new_vec
        return copied

    def __or__(self, other):
        """
        Makes one embedding orthogonal to the other one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo | bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} | {other.name})"
        copied.vector = self.vector - (self >> other).vector
        return copied

    def __repr__(self):
        return f"Emb[{self.name}]"

    def __str__(self):
        return self.name

    @property
    def norm(self):
        """Gives the norm of the vector of the embedding"""
        return np.linalg.norm(self.vector)

    def distance(self, other, metric: str = "cosine"):
        """
        Calculates the vector distance between two embeddings.

        Arguments:
            other: the other embedding you're comparing against
            metric: the distance metric to use, the list of valid options can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html)

        **Usage**

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [1.0, 0.0])
        bar = Embedding("bar", [0.0, 0.5])

        foo.distance(bar)
        foo.distance(bar, metric="euclidean")
        foo.distance(bar, metric="cosine")
        ```
        """
        return pairwise_distances([self.vector], [other.vector], metric=metric)[0][0]

    def plot(
        self,
        kind: str = "arrow",
        x_axis: Union[int, "Embedding"] = 0,
        y_axis: Union[int, "Embedding"] = 1,
        axis_metric: Optional[Union[str, Callable, Sequence]] = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        title: Optional[str] = None,
        color: str = None,
        show_ops: bool = False,
        annot: bool = True,
        axis_option: Optional[str] = None,
    ):
        """
        Handles the logic to perform a 2d plot in matplotlib.

        Arguments:
            kind: what kind of plot to make, can be `scatter`, `arrow` or `text`
            x_axis: the x-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            y_axis: the y-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            axis_metric: the metric used to project an embedding on the axes; only used when the corresponding
                axis (i.e. `x_axis` or `y_axis`) is an `Embedding` instance. It could be a string
                (`'cosine_similarity'`, `'cosine_distance'` or `'euclidean'`), or a callable that takes two vectors as input
                and returns a scalar value as output. To set different metrics for x- and y-axis, a list or a tuple of
                two elements could be given. By default (`None`), normalized scalar projection (i.e. `>` operator) is used.
            x_label: an optional label used for x-axis; if not given, it is set based on `x_axis` value.
            y_label: an optional label used for y-axis; if not given, it is set based on `y_axis` value.
            title: an optional title for the plot.
            color: the color of the dots
            show_ops: setting to also show the applied operations, only works for `text`
            annot: should the points be annotated
            axis_option: a string which is passed as `option` argument to `matplotlib.pyplot.axis` in order to control
                axis properties (e.g. using `'equal'` make circles shown circular in the plot). This might be useful
                for preserving geometric relationships (e.g. orthogonality) in the generated plot. See `matplotlib.pyplot.axis`
                [documentation](https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.axis.html#matplotlib-pyplot-axis)
                for possible values and their description.

        **Usage**
        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo.plot(kind="arrow", annot=True)
        bar.plot(kind="arrow", annot=True)
        ```
        """
        if isinstance(axis_metric, (list, tuple)):
            x_axis_metric = axis_metric[0]
            y_axis_metric = axis_metric[1]
        else:
            x_axis_metric = axis_metric
            y_axis_metric = axis_metric
        x_val, x_lab = self._get_plot_axis_value_and_label(
            x_axis, x_axis_metric, dir="x"
        )
        y_val, y_lab = self._get_plot_axis_value_and_label(
            y_axis, y_axis_metric, dir="y"
        )
        x_label = x_lab if x_label is None else x_label
        y_label = y_lab if y_label is None else y_label
        emb_plot = Embedding(name=self.name, vector=[x_val, y_val], orig=self.orig)
        handle_2d_plot(
            emb_plot,
            kind=kind,
            color=color,
            xlabel=x_label,
            ylabel=y_label,
            title=title,
            show_operations=show_ops,
            annot=annot,
            axis_option=axis_option,
        )
        return self

    def _get_plot_axis_value_and_label(self, axis, axis_metric, dir) -> tuple:
        """
        A helper function to get the projected value of this embedding on x- and y-axis,
        as well as the default label of axis.

        Arguments:
            axis: the axis value used for projection. It could be an integer or
                an `Embedding` instance.
            axis_metric: metric used for projection of embedding on axis (see `plot`
                method documentation).
            dir: the axis direction which could be either of `'x'` or `'y'`.
        """
        if isinstance(axis, int):
            return self.vector[axis], "Dimension " + str(axis)
        elif isinstance(axis, Embedding):
            if axis_metric is None:
                return self > axis, axis.name
            else:
                metric = self._get_plot_axis_metric_callable(axis_metric)
                return metric(self.vector, axis.vector), axis.name
        else:
            raise ValueError(
                f"The `{dir}_axis` value should be an integer or Embedding instance, given: {type(axis)}"
            )

    @staticmethod
    def _get_plot_axis_metric_callable(axis_metric) -> Optional[Callable]:
        """
        A helper function to retrieve a callable based on the given axis metric value.
        """
        if axis_metric is None:
            return None
        elif isinstance(axis_metric, str):
            if axis_metric == "cosine":
                raise ValueError(
                    "Please be specific, do you want `cosine_distance` or `cosine_similarity`?"
                )
            elif axis_metric == "cosine_similarity":
                return lambda x, y: 1.0 - scipy_distance.cosine(x, y)
            elif axis_metric == "cosine_distance":
                return scipy_distance.cosine
            elif axis_metric == "euclidean":
                return scipy_distance.euclidean
            else:
                raise ValueError(
                    f"The given axis metric is not supported: {axis_metric}"
                )
        elif callable(axis_metric):
            return axis_metric
        else:
            raise ValueError(
                f"The given axis metric type is not supported: {type(axis_metric)}"
            )



import numpy as np

from whatlies.language import TFHubLanguage, UniversalSentenceLanguage


def test_same_results():
    use_lang = UniversalSentenceLanguage("multi", 3)
    tf_lang = TFHubLanguage(
        "https://tfhub.dev/google/universal-sentence-encoder-multilingual/3"
    )
    assert np.allclose(use_lang["hello world"].vector, tf_lang["hello world"].vector)



import numpy as np
from sklearn.preprocessing import FunctionTransformer

from whatlies.transformers import Transformer
from whatlies import EmbeddingSet
from whatlies.transformers._common import new_embedding_dict


class Noise(Transformer):
    """
    This transformer adds gaussian noise to an embeddingset.

    Arguments:
        sigma: the amount of gaussian noise to add
        seed: seed value for random number generator

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Noise

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Noise(3))
    ```
    """

    def __init__(self, sigma=0.1, seed=42):
        super().__init__()
        self.seed = seed
        self.tfm = FunctionTransformer(
            lambda X: X + np.random.normal(0, sigma, X.shape)
        )

    def fit(self, embset):
        self.is_fitted = True
        return self

    def transform(self, embset):
        names, X = embset.to_names_X()
        np.random.seed(self.seed)
        new_vecs = self.tfm.transform(X)
        new_dict = new_embedding_dict(names, new_vecs, embset)
        return EmbeddingSet(
            new_dict,
            name=f"{embset.name}",
        )



from typing import Union

from ._tfhub_lang import TFHubLanguage


def UniversalSentenceLanguage(variant: str = "base", version: Union[int, None] = None):
    """
    Retreive a [universal sentence encoder](https://tfhub.dev/google/collections/universal-sentence-encoder/1) model from tfhub.

    You can download specific versions for specific variants. The variants that we support are listed below.

    - `"base"`: the base variant (915MB) [link](https://tfhub.dev/google/universal-sentence-encoder/4)
    - `"large"`: the large variant (523MB) [link](https://tfhub.dev/google/universal-sentence-encoder-large/5)
    - `"qa"`: the variant based on question/answer (528MB) [link](https://tfhub.dev/google/universal-sentence-encoder-qa/3)
    - `"multi"`: the multi-language variant (245MB) [link](https://tfhub.dev/google/universal-sentence-encoder-multilingual/3)
    - `"multi-large"`: the large multi-language variant (303MB) [link](https://tfhub.dev/google/universal-sentence-encoder-multilingual-large/3)
    - `"multi-qa"`: the multi-language qa variant (310MB) [link](https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/3)

    TFHub reports that the multi-language models support Arabic, Chinese-simplified, Chinese-traditional,
    English, French, German, Italian, Japanese, Korean, Dutch, Polish, Portuguese, Spanish, Thai, Turkish and Russian.

    Arguments:
        variant: select a specific variant
        version: select a specific version, if kept `None` we'll assume the most recent version
    """
    urls = {
        "base": "https://tfhub.dev/google/universal-sentence-encoder/",
        "large": "https://tfhub.dev/google/universal-sentence-encoder-large/",
        "qa": "https://tfhub.dev/google/universal-sentence-encoder-qa/",
        "multi": "https://tfhub.dev/google/universal-sentence-encoder-multilingual/",
        "multi-large": "https://tfhub.dev/google/universal-sentence-encoder-multilingual-large/",
        "multi-qa": "https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/3",
    }

    versions = {
        "base": 4,
        "large": 5,
        "qa": 3,
        "multi": 3,
        "multi-large": 3,
        "multi-qa": 3,
    }

    version = versions[variant] if not version else version
    url = urls[variant] + str(version)
    return TFHubLanguage(url=url)



import pytest

from whatlies.language import GensimLanguage


@pytest.fixture()
def lang():
    return GensimLanguage("tests/cache/custom_gensim_vectors.kv")


def test_missing_retreival(lang):
    assert lang["doesnotexist"].vector.shape == (10,)


def test_spaces_retreival(lang):
    assert lang["graph trees"].vector.shape == (10,)
    assert lang["graph trees dog"].vector.shape == (10,)


def test_single_token_words(lang):
    assert lang["computer"].vector.shape == (10,)
    assert len(lang[["red", "blue"]]) == 2


def test_similar_retreival(lang):
    assert len(lang.score_similar("hi", 10)) == 10
    assert len(lang.embset_similar("hi", 10)) == 10



from sklearn.preprocessing import normalize

from whatlies.transformers import Transformer
from whatlies import EmbeddingSet
from whatlies.transformers._common import new_embedding_dict


class Normalizer(Transformer):
    """
    This transformer normalizes the embeddings in an `EmbeddingSet` instance.

    Arguments:
        norm: the normalization value, could be either of `'l1'`, `'l2'` or `'max'`.
        feature: if `True` each feature (i.e. dimension) of embeddings would be normalized
            independently to unit norm; otheriwse, each embedding vector would be normazlied to unit norm.

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Normalizer

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Normalizer(norm='l1'))
    emb.transform(Normalizer(feature=True))
    ```
    """

    def __init__(self, norm: str = "l1", feature: bool = False) -> None:
        super().__init__()
        self.norm = norm
        self.feature = feature

    def fit(self, embset: EmbeddingSet) -> "Normalizer":
        self.is_fitted = True
        return self

    def transform(self, embset: EmbeddingSet) -> EmbeddingSet:
        names, X = embset.to_names_X()
        axis = 0 if self.feature else 1
        X = normalize(X, norm=self.norm, axis=axis)
        new_dict = new_embedding_dict(names, X, embset)
        return EmbeddingSet(new_dict, name=embset.name)



import pytest

import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.pipeline import FeatureUnion
from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import CountVectorizer

from whatlies.language import (
    FasttextLanguage,
    SpacyLanguage,
    GensimLanguage,
    BytePairLanguage,
    TFHubLanguage,
    HFTransformersLanguage,
)


backends = [
    SpacyLanguage("tests/custom_test_lang/"),
    FasttextLanguage("tests/custom_fasttext_model.bin"),
    BytePairLanguage("en", vs=1000, dim=25, cache_dir="tests/cache"),
    GensimLanguage("tests/cache/custom_gensim_vectors.kv"),
    HFTransformersLanguage("sshleifer/tiny-gpt2", framework="tf"),
    TFHubLanguage("https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim/1"),
]


@pytest.mark.parametrize("lang", backends)
def test_sklearn_pipeline_works(lang):
    pipe = Pipeline([("embed", lang), ("model", LogisticRegression())])

    X = [
        "i really like this post",
        "thanks for that comment",
        "i enjoy this friendly forum",
        "this is a bad post",
        "this is a bad post",
        "i dislike this article",
        "this is not well written",
    ]
    y = np.array([1, 1, 1, 0, 0, 0, 0])

    pipe.fit(X, y)
    assert pipe.predict(X).shape[0] == 7

    preprocess = FeatureUnion([("dense", lang), ("sparse", CountVectorizer())])

    assert preprocess.fit_transform(X).shape[0] == 7



import pytest
import numpy as np
from sklearn.utils import estimator_checks

from spacy.vocab import Vocab
from spacy.language import Language
from whatlies.language import SpacyLanguage


@pytest.fixture()
def color_lang():
    vector_data = {
        "red": np.array([1.0, 0.0]),
        "green": np.array([0.5, 0.5]),
        "blue": np.array([0.0, 1.0]),
        "purple": np.array([0.0, 1.0]),
    }

    vocab = Vocab(strings=vector_data.keys())
    for word, vector in vector_data.items():
        vocab.set_vector(word, vector)
    nlp = Language(vocab=vocab)
    return SpacyLanguage(nlp)


@pytest.mark.parametrize(
    "text", [("red red", "blue red"), ("red", "green", "blue"), ("dog", "cat")]
)
def test_check_sizes(color_lang, text):
    assert color_lang.fit(text).transform(text).shape == (len(text), 2)
    assert color_lang.fit_transform(text).shape == (len(text), 2)


def test_get_params():
    assert "nlp" in SpacyLanguage("tests/custom_test_lang/").get_params().keys()


checks = (
    estimator_checks.check_fit1d,
    estimator_checks.check_get_params_invariance,
    estimator_checks.check_set_params,
    estimator_checks.check_transformers_unfitted,
    # these checks won't work because they assume text data
    # estimator_checks.check_fit2d_predict1d,
    # estimator_checks.check_dont_overwrite_parameters,
    # estimator_checks.check_fit2d_1sample,
    # estimator_checks.check_fit2d_1feature,
    # estimator_checks.check_transformer_data_not_an_array,
    # estimator_checks.check_transformer_general,
    # estimator_checks.check_methods_subset_invariance,
    # estimator_checks.check_dict_unchanged,
)


@pytest.mark.parametrize("test_fn", checks)
def test_estimator_checks(test_fn):
    test_fn("spacy_lang", SpacyLanguage("tests/custom_test_lang/"))



from abc import ABC, abstractmethod

from whatlies import EmbeddingSet


class Transformer(ABC):
    """
    This is the abstract base class for all the transformers. Each subclass of
    this class should at least implement the `fit` and `transform` methods.
    """

    def __init__(self, *args, **kwargs) -> None:
        self.is_fitted = False

    def __call__(self, embset: EmbeddingSet) -> EmbeddingSet:
        if not self.is_fitted:
            self.fit(embset)
        return self.transform(embset)

    @abstractmethod
    def fit(self, embset: EmbeddingSet) -> "Transformer":
        """
        Fit the transformer on the given `EmbeddingSet` instance (if neccessary). This method should
        set the `is_fitted` flag to `True`.

        Arguments:
            embset: an `EmbeddingSet` instance used for fitting the transformer.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, embset: EmbeddingSet) -> EmbeddingSet:
        """
        Transform the given `EmbeddingSet` instance.

        Arguments:
            embset: an `EmbeddingSet` instance to be transformed.
        """
        raise NotImplementedError



def validate_plot_general_properties(chart_layer, props):
    assert chart_layer["encoding"]["x"]["axis"]["title"] == props["x_label"]
    assert chart_layer["encoding"]["y"]["axis"]["title"] == props["y_label"]
    assert chart_layer["title"] == props["title"]
    assert chart_layer["mark"]["type"] == props["type"]


def validate_matrix_plot_general_properties(spec, props):
    assert spec["width"] == props["width"]
    assert spec["height"] == props["height"]



import pytest
import numpy as np

from spacy.vocab import Vocab
from spacy.language import Language
from whatlies.language import SpacyLanguage


@pytest.fixture()
def color_lang():
    vector_data = {
        "red": np.array([1.0, 0.0]),
        "green": np.array([0.5, 0.5]),
        "blue": np.array([0.0, 1.0]),
        "purple": np.array([0.0, 1.0]),
    }

    vocab = Vocab(strings=vector_data.keys())
    for word, vector in vector_data.items():
        vocab.set_vector(word, vector)
    nlp = Language(vocab=vocab)
    return SpacyLanguage(nlp)


def test_basic_usage(color_lang):
    queries = [
        "green is blue and yellow",
        "purple is red and blue",
        "purple isn't same as red!",
        "[red and blue] is a like a glue!",
    ]
    emb = color_lang[queries]
    assert len(emb) == 4
    assert emb[queries[0]].name == "green is blue and yellow"
    assert emb[queries[0]].vector.shape == (2,)


@pytest.mark.parametrize(
    "query",
    [
        "no blue [closing bracket",
        "no red opening] bracket",
        "[more] than one blue [context]",
    ],
)
def test_invalid_query_raises_error(color_lang, query):
    with pytest.raises(ValueError, match="bracket"):
        color_lang[query]


@pytest.mark.parametrize(
    "query, context, context_pos",
    [
        ("I'm going [blue]", "blue", (2, 3)),
        ("you should pre-order your [red shirt]", "red shirt", (6, 8)),
        ("Me: today is the [best green day] of my life", "best green day", (5, 8)),
        ("[red and blue] is like a glue!", "red and blue", (0, 3)),
        (
            "Let's have Fun without a blue context.",
            "Let's have Fun without a blue context.",
            (0, None),
        ),
    ],
)
def test_embedding_is_correct(color_lang, query, context, context_pos):
    emb = color_lang[query]
    assert emb.vector.shape == (2,)
    assert emb.name == query

    clean_query = query.replace("[", "").replace("]", "")
    doc = color_lang.model(clean_query)
    span = doc[context_pos[0] : context_pos[1]]
    assert str(span) == context
    assert np.allclose(emb.vector, span.vector)


def test_score_similar_one(color_lang):
    scores = color_lang.score_similar("blue", n=2, prob_limit=None, lower=False)
    print(scores)
    assert all([s[1] == 0 for s in scores])
    assert "blue" in [s[0].name for s in scores]
    assert "purple" in [s[0].name for s in scores]


@pytest.mark.parametrize(
    "string, array",
    zip(
        ["red", "red green [blue] purple", "green [red blue] pink"],
        [np.array([1.0, 0.0]), np.array([0.0, 1.0]), np.array([0.5, 0.5])],
    ),
)
def test_lang_retreival(color_lang, string, array):
    assert np.isclose(color_lang[string].vector, array).all()


def test_single_token_words(color_lang):
    # test for issue here: https://github.com/RasaHQ/whatlies/issues/5
    assert len(color_lang["red"].vector) > 0


def test_raise_warning(color_lang):
    print([w for w in color_lang.nlp.vocab])
    with pytest.warns(UserWarning):
        color_lang.score_similar("red", 100, prob_limit=None, lower=False)



import pytest
import numpy as np
from transformers import AutoTokenizer

from whatlies.language import HFTransformersLanguage


def load_model_and_tokenizer(model_name, tensor_type):
    if tensor_type == "tf":
        from transformers import TFAutoModel as AutoModel
    elif tensor_type == "pt":
        from transformers import AutoModel

    model = AutoModel.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token
    return model, tokenizer


@pytest.mark.parametrize(
    "model_name, expected_shape",
    [
        ("sshleifer/tiny-gpt2", (2,)),
        ("sshleifer/tiny-distilroberta-base", (2,)),
    ],
)
@pytest.mark.parametrize("tensor_type", ["tf", "pt"])
def test_basic_usage_and_generated_embeddings(model_name, expected_shape, tensor_type):
    sentences = [
        "hello how are you?",
        "I'm fine, thanks!",
        "how about you?",
    ]

    lang = HFTransformersLanguage(model_name, framework=tensor_type)
    # Test for one single query
    emb = lang[sentences[0]]
    assert emb.vector.shape == expected_shape
    assert emb.name == sentences[0]

    model, tokenizer = load_model_and_tokenizer(model_name, tensor_type)
    tokens = tokenizer(
        sentences,
        padding=True,
        return_special_tokens_mask=True,
        return_tensors=tensor_type,
    )
    mask = tokens["special_tokens_mask"].numpy()
    del tokens["special_tokens_mask"]
    if tensor_type == "tf":
        output = model(tokens)
    elif tensor_type == "pt":
        output = model(**tokens)

    # Test for multiple queries
    emb = lang[sentences]
    assert len(emb) == len(sentences)
    for i, s in enumerate(sentences):
        if tensor_type == "tf":
            feats = output[0][i][np.logical_not(mask[i])].numpy()
        elif tensor_type == "pt":
            feats = output[0][i][np.logical_not(mask[i])].detach().numpy()
        assert emb[s].vector.shape == expected_shape
        assert np.allclose(emb[s].vector, feats.sum(axis=0))



from ._spacy_lang import SpacyLanguage
from ._fasttext_lang import FasttextLanguage
from ._countvector_lang import CountVectorLanguage
from ._bpemblang import BytePairLanguage
from ._gensim_lang import GensimLanguage
from ._convert_lang import ConveRTLanguage

from whatlies.error import NotInstalled

try:
    from ._tfhub_lang import TFHubLanguage
except ModuleNotFoundError as e:
    TFHubLanguage = NotInstalled("TFHubLanguage", "tfhub")

try:
    from ._sentence_encode_lang import UniversalSentenceLanguage
except ModuleNotFoundError as e:
    UniversalSentenceLanguage = NotInstalled("UniversalSentenceLanguage", "tfhub")

try:
    from ._hftransformers_lang import HFTransformersLanguage
except ModuleNotFoundError as e:
    HFTransformersLanguage = NotInstalled("HFTransformersLanguage", "transformers")

try:
    from ._sense2vec_lang import Sense2VecLanguage
except ModuleNotFoundError as e:
    Sense2VecLanguage = NotInstalled("Sense2VecLanguage", "sense2vec")


__all__ = [
    "SpacyLanguage",
    "Sense2VecLanguage",
    "FasttextLanguage",
    "CountVectorLanguage",
    "BytePairLanguage",
    "GensimLanguage",
    "ConveRTLanguage",
    "TFHubLanguage",
    "HFTransformersLanguage",
    "UniversalSentenceLanguage",
]



import pathlib
from setuptools import setup, find_packages

base_packages = [
    "numpy>=1.16.0",
    "scipy>=1.2.0",
    "scikit-learn>=0.20.2",
    "umap-learn>=0.3.10",
    "altair>=4.0.1",
    "matplotlib>=3.2.0",
    "spacy>=2.2.3",
    "spacy-lookups-data>=0.3.2",
    "networkx>=2.4",
    "fasttext>=0.9.1",
    "bpemb>=0.3.0",
    "gensim>=3.8.3",
]

s2v_packages = [
    "sense2vec>=1.0.2",
]

tf_packages = [
    "tensorflow>=2.3.0",
    "tensorflow-text>=2.3.0",
    "tensorflow-hub>=0.8.0",
]

transformers_dep = [
    "transformers>=3.0.0",
]

ivis_dep = [
    "ivis[cpu]>=1.8.0",
]

open_tsne_dep = [
    "opentsne>=0.4.3",
]

docs_packages = [
    "mkdocs==1.1",
    "mkdocs-material==4.6.3",
    "mkdocstrings==0.8.0",
    "jupyterlab>=0.35.4",
    "nbstripout>=0.3.7",
    "nbval>=0.9.5",
]

test_packages = [
    "torch>=1.6.0",
    "flake8>=3.6.0",
    "pytest>=4.0.2",
    "black>=19.3b0",
    "pytest-cov>=2.6.1",
    "nbval>=0.9.5",
    "pre-commit>=2.2.0",
]

extra_deps = tf_packages + transformers_dep + ivis_dep + open_tsne_dep + s2v_packages
dev_packages = docs_packages + test_packages + extra_deps


setup(
    name="whatlies",
    version="0.5.4",
    author="Vincent D. Warmerdam",
    packages=find_packages(exclude=["notebooks", "docs"]),
    description="Tools to help uncover `whatlies` in word embeddings.",
    long_description=pathlib.Path("readme.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://rasahq.github.io/whatlies/",
    project_urls={
        "Documentation": "https://rasahq.github.io/whatlies/",
        "Source Code": "https://github.com/RasaHQ/whatlies/",
        "Issue Tracker": "https://github.com/RasaHQ/whatlies/issues",
    },
    install_requires=base_packages,
    extras_require={
        "base": base_packages,
        "docs": docs_packages,
        "dev": dev_packages,
        "test": test_packages,
        "tfhub": tf_packages,
        "sense2vec": s2v_packages,
        "transformers": transformers_dep,
        "ivis": ivis_dep,
        "opentsne": open_tsne_dep,
        "all": extra_deps,
    },
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import pytest
from whatlies import Embedding, EmbeddingSet


def handle_docstring(doc, indent):
    """
    This function will read through the docstring and grab
    the first python code block. It will try to execute it.
    If it fails, the calling test should raise a flag.
    """
    if not doc:
        return
    start = doc.find("```python\n")
    end = doc.find("```\n")
    if start != -1:
        if end != -1:
            code_part = doc[(start + 10) : end].replace(" " * indent, "")
            print(code_part)
            exec(code_part)


@pytest.mark.parametrize("m", [m for m in dir(Embedding) if not m.startswith("_")])
def test_emb_docstrings(m):
    """
    Take the docstring of every method on the `Embedding` class.
    The test passes if the usage examples causes no errors.
    """
    handle_docstring(getattr(Embedding, m).__doc__, indent=8)


@pytest.mark.parametrize("m", [m for m in dir(EmbeddingSet) if not m.startswith("_")])
def test_embset_docstrings(m):
    """
    Take the docstring of every method on the `EmbeddingSet` class.
    The test passes if the usage examples causes no errors.
    """
    handle_docstring(getattr(EmbeddingSet, m).__doc__, indent=8)



import warnings
from pathlib import Path
from typing import Union, List

import numpy as np
from bpemb import BPEmb
from sklearn.metrics import pairwise_distances

from whatlies import Embedding, EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin


class BytePairLanguage(SklearnTransformerMixin):
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a Byte-Pair Encoding
    backend. This object is meant for retreival, not plotting.

    This language represents token-free pre-trained subword embeddings. Originally created by
    Benjamin Heinzerling and Michael Strube.

    Important:
        These vectors will auto-download by the [BPEmb package](https://nlp.h-its.org/bpemb/).
        You can also specify "multi" to download multi language embeddings. A full list of available
        languages can be found [here](https://nlp.h-its.org/bpemb). The article that
        belongs to this work can be found [here](http://www.lrec-conf.org/proceedings/lrec2018/pdf/1049.pdf)
        Recognition should be given to Benjamin Heinzerling and Michael Strube for making these available.
        The availability of vocabulary size as well as dimensionality can be varified
        on the project website. See [here](https://nlp.h-its.org/bpemb/en/) for an
        example link in English. Please credit the original authors if you use their work.

    Warning:
        This class used to be called `BytePairLang`.

    Arguments:
        lang: name of the model to load
        vs: vocabulary size of the byte pair model
        dim: the embedding dimensionality
        cache_dir: The folder in which downloaded BPEmb files will be cached

    Typically the vocabulary size given from this backend can be of size 1000,
    3000, 5000, 10000, 25000, 50000, 100000 or 200000. The available dimensionality
    of the embbeddings typically are 25, 50, 100, 200 and 300.

    **Usage**:

    ```python
    > from whatlies.language import BytePairLanguage
    > lang = BytePairLanguage(lang="en")
    > lang['python']
    > lang = BytePairLanguage(lang="multi")
    > lang[['hund', 'hond', 'dog']]
    ```
    """

    def __init__(
        self, lang, vs=10000, dim=100, cache_dir=Path.home() / Path(".cache/bpemb")
    ):
        self.module = BPEmb(lang=lang, vs=vs, dim=dim, cache_dir=cache_dir)

    def __getitem__(self, item):
        """
        Retreive a single embedding or a set of embeddings. If an embedding contains multiple
        sub-tokens then we'll average them before retreival.

        Arguments:
            item: single string or list of strings

        **Usage**
        ```python
        > lang = BytePairLanguage(lang="en")
        > lang['python']
        > lang[['python', 'snake']]
        > lang[['nobody expects', 'the spanish inquisition']]
        ```
        """
        if isinstance(item, str):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                return Embedding(item, self.module.embed(item).mean(axis=0))
        if isinstance(item, list):
            return EmbeddingSet(*[self[i] for i in item])
        raise ValueError(f"Item must be list of string got {item}.")

    def _prepare_queries(self, lower):
        queries = [w for w in self.module.emb.vocab.keys()]
        if lower:
            queries = [w for w in queries if w.lower() == w]
        return queries

    def _calculate_distances(self, emb, queries, metric):
        vec = emb.vector
        vector_matrix = np.array([self[w].vector for w in queries])
        # there are NaNs returned, good to investigate later why that might be
        vector_matrix = np.array(
            [np.zeros(v.shape) if np.any(np.isnan(v)) else v for v in vector_matrix]
        )
        return pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)

    def score_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        metric="cosine",
        lower=False,
    ) -> List:
        """
        Retreive a list of (Embedding, score) tuples that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(lower=lower)
        distances = self._calculate_distances(emb=emb, queries=queries, metric=metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])

        if len(queries) < n:
            warnings.warn(
                f"We could only find {len(queries)} feasible words. Consider changing `top_n` or `lower`",
                UserWarning,
            )

        return [(self[q], float(d)) for q, d in by_similarity[:n]]

    def embset_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        lower=False,
        metric="cosine",
    ) -> EmbeddingSet:
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Important:
            This method is incredibly slow at the moment without a good `top_n` setting due to
            [this bug](https://github.com/facebookresearch/fastText/issues/1040).

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [
            w[0] for w in self.score_similar(emb=emb, n=n, lower=lower, metric=metric)
        ]
        return EmbeddingSet({w.name: w for w in embs})



import numpy as np

from whatlies.transformers import Transformer
from whatlies import Embedding, EmbeddingSet


class AddRandom(Transformer):
    """
    This transformer adds random embeddings to the embeddingset.

    Arguments:
        n: the number of random vectors to add

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import AddRandom

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(AddRandom(3)).plot_interactive_matrix('rand_0', 'rand_1', 'rand_2')
    ```
    """

    def __init__(self, n=1, sigma=0.1, seed=42):
        super().__init__()
        self.n = n
        self.sigma = sigma
        self.seed = seed

    def fit(self, embset):
        self.is_fitted = True
        return self

    def transform(self, embset):
        names, X = embset.to_names_X()
        np.random.seed(self.seed)
        orig_dict = embset.embeddings.copy()
        new_dict = {
            f"rand_{k}": Embedding(
                f"rand_{k}", np.random.normal(0, self.sigma, X.shape[1])
            )
            for k in range(self.n)
        }
        return EmbeddingSet({**orig_dict, **new_dict})



import json

import pytest
import numpy as np
import pandas as pd
import scipy.spatial.distance as scipy_distance

from whatlies import Embedding, EmbeddingSet
from .common import validate_matrix_plot_general_properties


"""
*Guide*

Here are the plot's propertites which could be checked (some of them may not be applicable
for a particular plot/test case):
    - type: the type of plot; usually it's scatter plot with circle marks.
    - data: the position (i.e. coordinates) of datapoints in the plot.
    - x_repeat: the name of repeat field representing x-axis.
    - x_repeat_fields: the name of the fields of chart data which is used for x-axis coordinates.
    - y_repeat: the name of repeat field representing y-axis.
    - y_repeat_fields: the name of the fields of chart data which is used for y-axis coordinates.
    - label_field: the name of the field of chart data which is used for annotating data points with text labels.
    - label: the text labels used for annotation of datapoints.
    - width: the width of plot
    - height: the height of plot
"""


@pytest.fixture
def embset():
    names = ["red", "blue", "green", "yellow", "white"]
    vectors = np.random.rand(5, 4) * 10 - 5
    embeddings = [Embedding(name, vector) for name, vector in zip(names, vectors)]
    return EmbeddingSet(*embeddings)


def test_default(embset):
    p = embset.plot_interactive_matrix()
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data": embset.to_X()[:, :2],
        "x_repeat": "column",
        "x_repeat_fields": ["Dimension 0", "Dimension 1"],
        "y_repeat": "row",
        "y_repeat_fields": ["Dimension 1", "Dimension 0"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)

    # Check if it's an interactive plot (only done in this test case)
    assert "selection" in layers[0]
    # Check tooltip data (only done in this test case)
    tooltip_fields = set(
        [
            layers[0]["encoding"]["tooltip"][0]["field"],
            layers[0]["encoding"]["tooltip"][1]["field"],
        ]
    )
    assert tooltip_fields == set(["name", "original"])


def test_int_axis(embset):
    p = embset.plot_interactive_matrix(2, 3, 0, height=300)
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data": embset.to_X()[:, [2, 3, 0]],
        "x_repeat": "column",
        "x_repeat_fields": ["Dimension 2", "Dimension 3", "Dimension 0"],
        "y_repeat": "row",
        "y_repeat_fields": ["Dimension 0", "Dimension 3", "Dimension 2"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 300,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_int_axis_with_str_axes_metric(embset):
    p = embset.plot_interactive_matrix(3, 2, 1, axes_metric="euclidean", annot=False)
    chart = json.loads(p.to_json())
    props = {
        "type": "circle",
        "data": embset.to_X()[:, 3:0:-1],
        "x_repeat": "column",
        "x_repeat_fields": ["Dimension 3", "Dimension 2", "Dimension 1"],
        "y_repeat": "row",
        "y_repeat_fields": ["Dimension 1", "Dimension 2", "Dimension 3"],
        "height": 200,
        "width": 200,
        # Not applicabel: label_field, label
    }
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert chart["spec"]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert chart["spec"]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert chart["spec"]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert "text" not in chart["spec"]["encoding"]
    assert "layer" not in chart["spec"]
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_str_axis(embset):
    p = embset.plot_interactive_matrix("red", "blue", "green")
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e > embset["red"], e > embset["blue"], e > embset["green"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["red", "blue", "green"],
        "y_repeat": "row",
        "y_repeat_fields": ["green", "blue", "red"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_str_axis_with_str_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        "yellow", "blue", "green", axes_metric="cosine_similarity", width=350
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                1 - scipy_distance.cosine(e.vector, embset["yellow"].vector),
                1 - scipy_distance.cosine(e.vector, embset["blue"].vector),
                1 - scipy_distance.cosine(e.vector, embset["green"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["yellow", "blue", "green"],
        "y_repeat": "row",
        "y_repeat_fields": ["green", "blue", "yellow"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 350,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_str_axis_with_different_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        "yellow", "blue", "white", axes_metric=["cosine_distance", None, np.dot]
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                scipy_distance.cosine(e.vector, embset["yellow"].vector),
                e > embset["blue"],
                np.dot(e.vector, embset["white"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["yellow", "blue", "white"],
        "y_repeat": "row",
        "y_repeat_fields": ["white", "blue", "yellow"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_emb_axis(embset):
    p = embset.plot_interactive_matrix(embset["blue"], embset["green"])
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e > embset["blue"], e > embset["green"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["blue", "green"],
        "y_repeat": "row",
        "y_repeat_fields": ["green", "blue"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_emb_axis_with_str_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        embset["red"], embset["green"], embset["blue"], axes_metric="euclidean"
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                scipy_distance.euclidean(e.vector, embset["red"].vector),
                scipy_distance.euclidean(e.vector, embset["green"].vector),
                scipy_distance.euclidean(e.vector, embset["blue"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["red", "green", "blue"],
        "y_repeat": "row",
        "y_repeat_fields": ["blue", "green", "red"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_emb_axis_with_different_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        embset["red"],
        embset["green"],
        embset["blue"],
        axes_metric=[None, "cosine_distance", np.dot],
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                e > embset["red"],
                scipy_distance.cosine(e.vector, embset["green"].vector),
                np.dot(e.vector, embset["blue"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["red", "green", "blue"],
        "y_repeat": "row",
        "y_repeat_fields": ["blue", "green", "red"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_mixed_axis(embset):
    p = embset.plot_interactive_matrix("blue", 2, embset["red"])
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append([e > embset["blue"], e.vector[2], e > embset["red"]])
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["blue", "Dimension 2", "red"],
        "y_repeat": "row",
        "y_repeat_fields": ["red", "Dimension 2", "blue"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_mixed_axis_with_str_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        "blue", 2, embset["red"], axes_metric="cosine_similarity"
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                1 - scipy_distance.cosine(e.vector, embset["blue"].vector),
                e.vector[2],
                1 - scipy_distance.cosine(e.vector, embset["red"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["blue", "Dimension 2", "red"],
        "y_repeat": "row",
        "y_repeat_fields": ["red", "Dimension 2", "blue"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_mixed_axis_with_different_axes_metric(embset):
    p = embset.plot_interactive_matrix(
        "blue", embset["red"], 3, axes_metric=[np.dot, None, "cosine_distance"]
    )
    chart = json.loads(p.to_json())
    vectors = []
    for e in embset.embeddings.values():
        vectors.append(
            [
                np.dot(e.vector, embset["blue"].vector),
                e > embset["red"],
                e.vector[3],
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": "circle",
        "data": vectors,
        "x_repeat": "column",
        "x_repeat_fields": ["blue", "red", "Dimension 3"],
        "y_repeat": "row",
        "y_repeat_fields": ["Dimension 3", "red", "blue"],
        "label_field": "original",
        "label": [v.orig for v in embset.embeddings.values()],
        "height": 200,
        "width": 200,
    }
    layers = chart["spec"]["layer"]
    chart_data = pd.DataFrame(chart["datasets"][chart["spec"]["data"]["name"]])
    assert layers[0]["mark"] == props["type"]
    assert np.array_equal(chart_data[props["x_repeat_fields"]].values, props["data"])
    assert layers[0]["encoding"]["x"]["field"]["repeat"] == props["x_repeat"]
    assert layers[0]["encoding"]["y"]["field"]["repeat"] == props["y_repeat"]
    assert chart["repeat"]["column"] == props["x_repeat_fields"]
    assert chart["repeat"]["row"] == props["y_repeat_fields"]
    assert layers[1]["encoding"]["text"]["field"] == props["label_field"]
    assert np.array_equal(chart_data[props["label_field"]], props["label"])
    validate_matrix_plot_general_properties(chart["spec"], props)


def test_raises_error_when_number_of_axes_is_not_equal_to_number_of_metrics(embset):
    with pytest.raises(
        ValueError, match="The number of given axes metrics should be the same"
    ):
        embset.plot_interactive_matrix("red", 3, 1, axes_metric=["euclidean", np.dot])



class ConveRTLanguage:
    """
    Important:
        This model has been deprecated. The original authors took the embeddings down.
    """

    MODEL_URL = {
        "convert": "https://github.com/PolyAI-LDN/polyai-models/releases/download/v1.0/model.tar.gz",
        "convert-multi-context": "https://github.com/PolyAI-LDN/polyai-models/releases/download/v1.0/model_multicontext.tar.gz",
        "convert-ubuntu": "https://github.com/PolyAI-LDN/polyai-models/releases/download/v1.0/model_ubuntu.tar.gz",
    }

    MODEL_SIGNATURES = [
        "default",
        "encode_context",
        "encode_response",
        "encode_sequence",
    ]

    def __init__(self, model_id: str = "convert", signature: str = "default") -> None:
        pass

    def __getitem__(self, item):
        raise DeprecationWarning(
            "This model has been deprecated. The original authors took the embeddings down."
        )



from copy import deepcopy
from functools import reduce
from collections import Counter
from typing import Union, Optional, Callable, Sequence, List

import numpy as np
import pandas as pd
import matplotlib.pylab as plt
import altair as alt
from sklearn.utils import deprecated
from sklearn.preprocessing import normalize
from sklearn.metrics import pairwise_distances
from sklearn.metrics.pairwise import (
    paired_distances,
    cosine_similarity,
    cosine_distances,
    euclidean_distances,
)

from whatlies.embedding import Embedding
from whatlies.common import plot_graph_layout, handle_2d_plot


class EmbeddingSet:
    """
    This object represents a set of `Embedding`s. You can use the same operations
    as an `Embedding` but here we apply it to the entire set instead of a single
    `Embedding`.

    **Parameters**

    - **embeddings**: list of `Embedding`, or a single dictionary containing name:`Embedding` pairs
    - **name**: custom name of embeddingset

    Usage:

    ```python
    from whatlies.embedding import Embedding
    from whatlies.embeddingset import EmbeddingSet

    foo = Embedding("foo", [0.1, 0.3])
    bar = Embedding("bar", [0.7, 0.2])
    emb = EmbeddingSet(foo, bar)
    emb = EmbeddingSet({'foo': foo, 'bar': bar)
    ```
    """

    def __init__(self, *embeddings, name=None):
        if not name:
            name = "EmbSet"
        self.name = name
        if isinstance(embeddings[0], dict):
            # Assume it's a single dictionary.
            self.embeddings = embeddings[0]
        else:
            # Assume it's a list of `Embedding` instances.
            names = [t.name for t in embeddings]
            if len(names) != len(set(names)):
                double_names = [k for k, v in Counter(names).items() if v > 1]
                raise Warning(
                    f"Some embeddings given to `EmbeddingSet` have the same name: {double_names}."
                )
            self.embeddings = {t.name: t for t in embeddings}

        # We cannot allow for different shapes because that will break many operations later.
        uniq_shapes = set(v.vector.shape for k, v in self.embeddings.items())
        if len(uniq_shapes) > 1:
            raise ValueError("Not all vectors have the same shape.")

    @property
    def ndim(self):
        """
        Return dimension of embedding vectors in embeddingset.
        """
        return next(iter(self.embeddings.values())).ndim

    def __contains__(self, item):
        """
        Checks if an item is in the embeddingset.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        "foo" in emb # True
        "dinosaur" in emb # False
        ```
        """
        return item in self.embeddings.keys()

    def __iter__(self):
        """
        Iterate over all the embeddings in the embeddingset.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        [e for e in emb]
        ```
        """
        return self.embeddings.values().__iter__()

    def __add__(self, other):
        """
        Adds an embedding to each element in the embeddingset.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        (emb).plot(kind="arrow")
        (emb + buz).plot(kind="arrow")
        ```
        """
        new_embeddings = {k: emb + other for k, emb in self.embeddings.items()}
        return EmbeddingSet(new_embeddings, name=f"({self.name} + {other.name})")

    def __sub__(self, other):
        """
        Subtracts an embedding from each element in the embeddingset.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        (emb).plot(kind="arrow")
        (emb - buz).plot(kind="arrow")
        ```
        """
        new_embeddings = {k: emb - other for k, emb in self.embeddings.items()}
        return EmbeddingSet(new_embeddings, name=f"({self.name} - {other.name})")

    def __or__(self, other):
        """
        Makes every element in the embeddingset othogonal to the passed embedding.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        (emb).plot(kind="arrow")
        (emb | buz).plot(kind="arrow")
        ```
        """
        new_embeddings = {k: emb | other for k, emb in self.embeddings.items()}
        return EmbeddingSet(new_embeddings, name=f"({self.name} | {other.name})")

    def __rshift__(self, other):
        """
        Maps every embedding in the embedding set unto the passed embedding.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar)

        (emb).plot(kind="arrow")
        (emb >> buz).plot(kind="arrow")
        ```
        """
        new_embeddings = {k: emb >> other for k, emb in self.embeddings.items()}
        return EmbeddingSet(new_embeddings, name=f"({self.name} >> {other.name})")

    def compare_against(
        self, other: Union[str, Embedding], mapping: Optional[Callable] = None
    ) -> List:
        """
        Compare (or map) the embeddigns in the embeddingset to a given embedding, optionally using
        a custom mapping function.

        Arguments:
            other: an `Embedding` instance, or name of an existing embedding; it is used for
                comparison with each embedding in the embeddingset.
            mapping: an optional callable used for for comparison that takes two 1D vector arrays as
                input; if not given, the normalized scalar projection (i.e. `>` operator) is used.
        """
        if isinstance(other, str):
            other = self[other]
        if mapping is None:
            return [v > other for v in self.embeddings.values()]
        elif callable(mapping):
            return [mapping(v.vector, other.vector) for v in self.embeddings.values()]
        else:
            raise ValueError(f"Unrecognized mapping value/type, got: {mapping}")

    def pipe(self, func, *args, **kwargs):
        """
        Applies a function to the embedding set. Useful for method chaining and
        chunks of code that repeat.

        Arguments:
             func: callable that accepts an `EmbeddingSet` set as its first argument
             args: arguments to also pass to the function
             kwargs: keyword arguments to also pass to the function

        ```python
        from whatlies.language import SpacyLanguage, BytePairLanguage

        lang_sp = SpacyLanguage("en_core_web_sm")
        lang_bp = BytePairLanguage("en", dim=25, vs=1000)

        text = ["cat", "dog", "rat", "blue", "red", "yellow"]

        def make_plot(embset):
            return (embset
                    .plot_interactive("dog", "blue")
                    .properties(height=200, width=200))

        p1 = lang_sp[text].pipe(make_plot)
        p2 = lang_bp[text].pipe(make_plot)
        p1 | p2
        ```
        """
        return func(self, *args, **kwargs)

    def to_X(self, norm=False):
        """
        Takes every vector in each embedding and turns it into a scikit-learn compatible `X` matrix.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar, buz)

        X = emb.to_X()
        ```
        """
        X = np.array([i.vector for i in self.embeddings.values()])
        X = normalize(X) if norm else X
        return X

    def to_X_y(self, y_label):
        """
        Takes every vector in each embedding and turns it into a scikit-learn compatible `X` matrix.
        Also retreives an array with potential labels.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        bla = Embedding("bla", [0.2, 0.8])

        emb1 = EmbeddingSet(foo, bar).add_property("label", lambda d: 'group-one')
        emb2 = EmbeddingSet(buz, bla).add_property("label", lambda d: 'group-two')
        emb = emb1.merge(emb2)

        X, y = emb.to_X_y(y_label='label')
        ```
        """
        X = self.to_X()
        y = np.array([getattr(e, y_label) for e in self.embeddings.values()])
        return X, y

    def to_names_X(self):
        """
        Get the list of names as well as an array of vectors of all embeddings.

        Usage:

        ```python
        from whatlies.embedding import Embedding
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])
        buz = Embedding("buz", [0.1, 0.9])
        emb = EmbeddingSet(foo, bar, buz)

        names, X = emb.to_names_X()
        ```
        """
        return list(self.embeddings.keys()), self.to_X()

    @classmethod
    def from_names_X(cls, names, X):
        """
        Constructs an `EmbeddingSet` instance from the given embedding names and vectors.

        Arguments:
            names: an iterable containing the names of embeddings
            X: an iterable of 1D vectors, or a 2D numpy array; it should have the same length as `names`

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet

        names = ["foo", "bar", "buz"]
        vecs = [
            [0.1, 0.3],
            [0.7, 0.2],
            [0.1, 0.9],
        ]

        emb = EmbeddingSet.from_names_X(names, vecs)
        ```
        """
        X = np.array(X)
        if len(X) != len(names):
            raise ValueError(
                f"The number of given names ({len(names)}) and vectors ({len(X)}) should be the same."
            )
        return cls({n: Embedding(n, v) for n, v in zip(names, X)})

    def transform(self, transformer):
        """
        Applies a transformation on the entire set.

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet
        from whatlies.transformers import Pca

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        buz = Embedding("buz", [0.1, 0.9, 0.12])
        emb = EmbeddingSet(foo, bar, buz).transform(Pca(2))
        ```
        """
        return transformer(self)

    def __getitem__(self, thing):
        """
        Retreive a single embedding from the embeddingset.

        Usage:
        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        buz = Embedding("buz", [0.1, 0.9, 0.12])
        emb = EmbeddingSet(foo, bar, buz)

        emb["buz"]
        ```
        """
        if isinstance(thing, str):
            return self.embeddings[thing]
        new_embeddings = {t: self[t] for t in thing}
        names = ",".join(thing)
        return EmbeddingSet(new_embeddings, name=f"{self.name}.subset({names})")

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __len__(self):
        return len(self.embeddings.keys())

    def filter(self, func):
        """
        Filters the collection of embeddings based on a predicate function.

        Arguments:
             func: callable that accepts a single embedding and outputs a boolean

        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        buz = Embedding("buz", [0.1, 0.9, 0.12])
        xyz = Embedding("xyz", [0.1, 0.9, 0.12])
        emb = EmbeddingSet(foo, bar, buz, xyz)
        emb.filter(lambda e: "foo" not in e.name)
        ```
        """
        return EmbeddingSet({k: v for k, v in self.embeddings.items() if func(v)})

    def merge(self, other):
        """
        Concatenates two embeddingssets together

        Arguments:
            other: another embeddingset

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        buz = Embedding("buz", [0.1, 0.9, 0.12])
        xyz = Embedding("xyz", [0.1, 0.9, 0.12])
        emb1 = EmbeddingSet(foo, bar)
        emb2 = EmbeddingSet(xyz, buz)

        both = emb1.merge(emb2)
        ```
        """
        return EmbeddingSet({**self.embeddings, **other.embeddings})

    def assign(self, **kwargs):
        """
        Adds properties to every embedding in the set based on the keyword arguments.

        This is very useful for plotting because a property can be used to assign colors. This method is very
        similar to `.add_property` but it might be more convenient when you want to assign multiple properties
        in one single statement.

        Arguments:
            kwargs: (name, func)-pairs that describe the name of the property as well as a function on how to calculate it.
                The function expects an `Embedding` object as input.

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        emb = EmbeddingSet(foo, bar)
        emb_with_property = emb.assign(dim0=lambda d: d.vector[0],
                                       dim1=lambda d: d.vector[1],
                                       dim2=lambda d: d.vector[2])
        ```
        """
        new_set = {}
        for k, e in self.embeddings.items():
            new_emb = e
            for name, func in kwargs.items():
                new_emb = new_emb.add_property(name, func)
            new_set[k] = new_emb
        return EmbeddingSet(new_set)

    def add_property(self, name, func):
        """
        Adds a property to every embedding in the set. Very useful for plotting because
        a property can be used to assign colors.

        Arguments:
            name: name of the property to add
            func: function that receives an embedding and needs to output the property value

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [0.1, 0.3, 0.10])
        bar = Embedding("bar", [0.7, 0.2, 0.11])
        emb = EmbeddingSet(foo, bar)
        emb_with_property = emb.add_property('example', lambda d: 'group-one')
        ```
        """
        return EmbeddingSet(
            {k: e.add_property(name, func) for k, e in self.embeddings.items()}
        )

    def average(self, name=None):
        """
        Takes the average over all the embedding vectors in the embeddingset. Turns it into
        a new `Embedding`.

        Arguments:
            name: manually specify the name of the average embedding

        Usage:

        ```python
        from whatlies.embeddingset import EmbeddingSet

        foo = Embedding("foo", [1.0, 0.0])
        bar = Embedding("bar", [0.0, 1.0])
        emb = EmbeddingSet(foo, bar)

        emb.average().vector                   # [0.5, 0,5]
        emb.average(name="the-average").vector # [0.5, 0.5]
        ```
        """
        name = f"{self.name}.average()" if not name else name
        x = self.to_X()
        return Embedding(name, np.mean(x, axis=0))

    def embset_similar(self, emb: Union[str, Embedding], n: int = 10, metric="cosine"):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most simmilar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [w[0] for w in self.score_similar(emb, n, metric)]
        return EmbeddingSet({w.name: w for w in embs})

    def score_similar(self, emb: Union[str, Embedding], n: int = 10, metric="cosine"):
        """
        Retreive a list of (Embedding, score) tuples that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if n > len(self):
            raise ValueError(
                f"You cannot retreive (n={n}) more items than exist in the Embeddingset (len={len(self)})"
            )

        if isinstance(emb, str):
            if emb not in self.embeddings.keys():
                raise ValueError(
                    f"Embedding for `{emb}` does not exist in this EmbeddingSet"
                )
            emb = self[emb]

        vec = emb.vector
        queries = [w for w in self.embeddings.keys()]
        vector_matrix = self.to_X()
        distances = pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])
        return [(self[q], float(d)) for q, d in by_similarity[:n]]

    def to_matrix(self):
        """
        Does exactly the same as `.to_X`. It takes the embedding vectors and turns it into a numpy array.
        """
        return self.to_X()

    def to_dataframe(self):
        """
        Turns the embeddingset into a pandas dataframe.
        """
        mat = self.to_matrix()
        return pd.DataFrame(mat, index=list(self.embeddings.keys()))

    def movement_df(self, other, metric="euclidean"):
        """
        Creates a dataframe that shows the movement from one embeddingset to another one.

        Arguments:
            other: the other embeddingset to compare against, will only keep the overlap
            metric: metric to use to calculate movement, must be scipy or sklearn compatible

        Usage:

        ```python
        from whatlies.language import SpacyLanguage

        lang = SpacyLanguage("en_core_web_sm")

        names = ['red', 'blue', 'green', 'yellow', 'cat', 'dog', 'mouse', 'rat', 'bike', 'car']
        emb = lang[names]
        emb_ort = lang[names] | lang['cat']
        emb.movement_df(emb_ort)
        ```
        """
        overlap = list(
            set(self.embeddings.keys()).intersection(set(other.embeddings.keys()))
        )
        mat1 = np.array([w.vector for w in self[overlap]])
        mat2 = np.array([w.vector for w in other[overlap]])
        return (
            pd.DataFrame(
                {"name": overlap, "movement": paired_distances(mat1, mat2, metric)}
            )
            .sort_values(["movement"], ascending=False)
            .reset_index()
        )

    def to_axis_df(self, x_axis, y_axis):
        if isinstance(x_axis, str):
            x_axis = self[x_axis]
        if isinstance(y_axis, str):
            y_axis = self[y_axis]
        return pd.DataFrame(
            {
                "x_axis": self.compare_against(x_axis),
                "y_axis": self.compare_against(y_axis),
                "name": [v.name for v in self.embeddings.values()],
                "original": [v.orig for v in self.embeddings.values()],
            }
        )

    def plot(
        self,
        kind: str = "arrow",
        x_axis: Union[int, str, Embedding] = 0,
        y_axis: Union[int, str, Embedding] = 1,
        axis_metric: Optional[Union[str, Callable, Sequence]] = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        title: Optional[str] = None,
        color: str = None,
        show_ops: bool = False,
        annot: bool = True,
        axis_option: Optional[str] = None,
    ):
        """
        Makes (perhaps inferior) matplotlib plot. Consider using `plot_interactive` instead.

        Arguments:
            kind: what kind of plot to make, can be `scatter`, `arrow` or `text`
            x_axis: the x-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            y_axis: the y-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            axis_metric: the metric used to project each embedding on the axes; only used when the corresponding
                axis (i.e. `x_axis` or `y_axis`) is a string or an `Embedding` instance. It could be a string
                (`'cosine_similarity'`, `'cosine_distance'` or `'euclidean'`), or a callable that takes two vectors as input
                and returns a scalar value as output. To set different metrics for x- and y-axis, a list or a tuple of
                two elements could be given. By default (`None`), normalized scalar projection (i.e. `>` operator) is used.
            x_label: an optional label used for x-axis; if not given, it is set based on value of `x_axis`.
            y_label: an optional label used for y-axis; if not given, it is set based on value of `y_axis`.
            title: an optional title for the plot.
            color: the color of the dots
            show_ops: setting to also show the applied operations, only works for `text`
            annot: should the points be annotated
            axis_option: a string which is passed as `option` argument to `matplotlib.pyplot.axis` in order to control
                axis properties (e.g. using `'equal'` make circles shown circular in the plot). This might be useful
                for preserving geometric relationships (e.g. orthogonality) in the generated plot. See `matplotlib.pyplot.axis`
                [documentation](https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.axis.html#matplotlib-pyplot-axis)
                for possible values and their description.
        """
        if isinstance(x_axis, str):
            x_axis = self[x_axis]
        if isinstance(y_axis, str):
            y_axis = self[y_axis]

        if isinstance(axis_metric, (list, tuple)):
            x_axis_metric = axis_metric[0]
            y_axis_metric = axis_metric[1]
        else:
            x_axis_metric = axis_metric
            y_axis_metric = axis_metric

        embeddings = []
        for emb in self.embeddings.values():
            x_val, x_lab = emb._get_plot_axis_value_and_label(
                x_axis, x_axis_metric, dir="x"
            )
            y_val, y_lab = emb._get_plot_axis_value_and_label(
                y_axis, y_axis_metric, dir="y"
            )
            emb_plot = Embedding(name=emb.name, vector=[x_val, y_val], orig=emb.orig)
            embeddings.append(emb_plot)
        x_label = x_lab if x_label is None else x_label
        y_label = y_lab if y_label is None else y_label
        handle_2d_plot(
            embeddings,
            kind=kind,
            color=color,
            xlabel=x_label,
            ylabel=y_label,
            title=title,
            show_operations=show_ops,
            annot=annot,
            axis_option=axis_option,
        )
        return self

    def plot_3d(
        self,
        x_axis: Union[int, str, Embedding] = 0,
        y_axis: Union[int, str, Embedding] = 1,
        z_axis: Union[int, str, Embedding] = 2,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        z_label: Optional[str] = None,
        title: Optional[str] = None,
        color: str = None,
        axis_metric: Optional[Union[str, Callable, Sequence]] = None,
        annot: bool = True,
    ):
        """
        Creates a 3d visualisation of the embedding.

        Arguments:
            x_axis: the x-axis to be used, must be given when dim > 3; if an integer, the corresponding
                dimension of embedding is used.
            y_axis: the y-axis to be used, must be given when dim > 3; if an integer, the corresponding
                dimension of embedding is used.
            z_axis: the z-axis to be used, must be given when dim > 3; if an integer, the corresponding
                dimension of embedding is used.
            x_label: an optional label used for x-axis; if not given, it is set based on value of `x_axis`.
            y_label: an optional label used for y-axis; if not given, it is set based on value of `y_axis`.
            z_label: an optional label used for z-axis; if not given, it is set based on value of `z_axis`.
            title: an optional title for the plot.
            color: the property to user for the color
            axis_metric: the metric used to project each embedding on the axes; only used when the corresponding
                axis is a string or an `Embedding` instance. It could be a string (`'cosine_similarity'`,
                `'cosine_distance'` or `'euclidean'`), or a callable that takes two vectors as input and
                returns a scalar value as output. To set different metrics of the three different axes,
                you can pass a list/tuple of size three that describes the metrics you're interested in.
                By default (`None`), normalized scalar projection (i.e. `>` operator) is used.
            annot: drawn points should be annotated

        **Usage**

        ```python
        from whatlies.language import SpacyLanguage
        from whatlies.transformers import Pca

        words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
                 "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
                 "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
                 "person", "family", "brother", "sister"]

        lang = SpacyLanguage("en_core_web_sm")
        emb = lang[words]

        emb.transform(Pca(3)).plot_3d(annot=True)
        emb.transform(Pca(3)).plot_3d("king", "dog", "red")
        emb.transform(Pca(3)).plot_3d("king", "dog", "red", axis_metric="cosine_distance")
        ```
        """
        if isinstance(x_axis, str):
            x_axis = self[x_axis]
        if isinstance(y_axis, str):
            y_axis = self[y_axis]
        if isinstance(z_axis, str):
            z_axis = self[z_axis]

        if isinstance(axis_metric, (list, tuple)):
            x_axis_metric = axis_metric[0]
            y_axis_metric = axis_metric[1]
            z_axis_metric = axis_metric[2]
        else:
            x_axis_metric = axis_metric
            y_axis_metric = axis_metric
            z_axis_metric = axis_metric

        # Determine axes values and labels
        if isinstance(x_axis, int):
            x_val = self.to_X()[:, x_axis]
            x_lab = "Dimension " + str(x_axis)
        else:
            x_axis_metric = Embedding._get_plot_axis_metric_callable(x_axis_metric)
            x_val = self.compare_against(x_axis, mapping=x_axis_metric)
            x_lab = x_axis.name
        x_lab = x_label if x_label is not None else x_lab

        if isinstance(y_axis, int):
            y_val = self.to_X()[:, y_axis]
            y_lab = "Dimension " + str(y_axis)
        else:
            y_axis_metric = Embedding._get_plot_axis_metric_callable(y_axis_metric)
            y_val = self.compare_against(y_axis, mapping=y_axis_metric)
            y_lab = y_axis.name
        y_lab = y_label if y_label is not None else y_lab

        if isinstance(z_axis, int):
            z_val = self.to_X()[:, z_axis]
            z_lab = "Dimension " + str(z_axis)
        else:
            z_axis_metric = Embedding._get_plot_axis_metric_callable(z_axis_metric)
            z_val = self.compare_against(z_axis, mapping=z_axis_metric)
            z_lab = z_axis.name
        z_lab = z_label if z_label is not None else z_lab

        # Save relevant information in a dataframe for plotting later.
        plot_df = pd.DataFrame(
            {
                "x_axis": x_val,
                "y_axis": y_val,
                "z_axis": z_val,
                "name": [v.name for v in self.embeddings.values()],
                "original": [v.orig for v in self.embeddings.values()],
            }
        )

        # Deal with the colors of the dots.
        if color:
            plot_df["color"] = [
                getattr(v, color) if hasattr(v, color) else ""
                for v in self.embeddings.values()
            ]

            color_map = {k: v for v, k in enumerate(set(plot_df["color"]))}
            color_val = [
                color_map[k] if not isinstance(k, float) else k
                for k in plot_df["color"]
            ]
        else:
            color_val = None

        ax = plt.axes(projection="3d")
        ax.scatter3D(
            plot_df["x_axis"], plot_df["y_axis"], plot_df["z_axis"], c=color_val, s=25
        )

        # Set the labels, titles, text annotations.
        ax.set_xlabel(x_lab)
        ax.set_ylabel(y_lab)
        ax.set_zlabel(z_lab)

        if annot:
            for i, row in plot_df.iterrows():
                ax.text(
                    row["x_axis"], row["y_axis"], row["z_axis"] + 0.05, row["original"]
                )
        if title:
            ax.set_title(label=title)
        return ax

    def plot_graph_layout(self, kind="cosine", **kwargs):
        plot_graph_layout(self.embeddings, kind, **kwargs)
        return self

    @deprecated(
        "This method will be deprecated in v0.6.0 in favor of `plot_distance` and `plot_similarity`"
    )
    def plot_correlation(self, metric=None):
        """
        Make a correlation plot. Shows you the correlation between all the word embeddings. Can
        also be configured to show distances instead.

        Arguments:
            metric: don't plot correlation but a distance measure, must be scipy compatible (cosine, euclidean, etc)

        Warning:
            This method will be deprecated in version 0.6.0 in favor of `plot_distance` and `plot_similarity`.

        Usage:

        ```python
        from whatlies.language import SpacyLanguage
        import matplotlib.pyplot as plt

        lang = SpacyLanguage("en_core_web_sm")

        names = ['red', 'blue', 'green', 'yellow', 'cat', 'dog', 'mouse', 'rat', 'bike', 'car']
        emb = lang[names]
        emb.plot_correlation()
        ```
        """
        df = self.to_dataframe().T
        corr_df = (
            pairwise_distances(self.to_matrix(), metric=metric) if metric else df.corr()
        )

        fig, ax = plt.subplots()
        plt.imshow(corr_df)
        plt.xticks(range(len(df.columns)), df.columns)
        plt.yticks(range(len(df.columns)), df.columns)
        plt.colorbar()

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")

    def plot_similarity(self, metric="cosine", norm=False):
        """
        Make a similarity plot. Shows you the similarity between all the word embeddings in the set.

        Arguments:
            metric: `'cosine'` or `'correlation'`
            norm: normalise the embeddings before calculating the similarity

        Usage:

        ```python
        from whatlies.language import SpacyLanguage
        lang = SpacyLanguage("en_core_web_sm")

        names = ['red', 'blue', 'green', 'yellow', 'cat', 'dog', 'mouse', 'rat', 'bike', 'car']
        emb = lang[names]
        emb.plot_similarity()
        emb.plot_similarity(metric='correlation')
        ```
        """
        allowed_metrics = ["cosine", "correlation"]
        if metric not in allowed_metrics:
            raise ValueError(
                f"The `metric` argument must be in {allowed_metrics}, got: {metric}."
            )

        vmin, vmax = 0, 1
        X = self.to_X(norm=norm)
        if metric == "cosine":
            similarity = cosine_similarity(X)
        if metric == "correlation":
            similarity = np.corrcoef(X)
            vmin, vmax = -1, 1

        fig, ax = plt.subplots()
        plt.imshow(similarity, cmap=plt.cm.get_cmap(), vmin=-vmin, vmax=vmax)
        plt.xticks(range(len(self)), self.embeddings.keys())
        plt.yticks(range(len(self)), self.embeddings.keys())
        plt.colorbar()

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")

    def plot_distance(self, metric="cosine", norm=False):
        """
        Make a distance plot. Shows you the distance between all the word embeddings in the set.

        Arguments:
            metric: `'cosine'`, `'correlation'` or `'euclidean'`
            norm: normalise the vectors before calculating the distances

        Usage:

        ```python
        from whatlies.language import SpacyLanguage
        lang = SpacyLanguage("en_core_web_sm")

        names = ['red', 'blue', 'green', 'yellow', 'cat', 'dog', 'mouse', 'rat', 'bike', 'car']
        emb = lang[names]
        emb.plot_distance(metric='cosine')
        emb.plot_distance(metric='euclidean')
        emb.plot_distance(metric='correlation')
        ```
        """
        allowed_metrics = ["cosine", "correlation", "euclidean"]
        if metric not in allowed_metrics:
            raise ValueError(
                f"The `metric` argument must be in {allowed_metrics}, got: {metric}."
            )

        vmin, vmax = 0, 1
        X = self.to_X(norm=norm)
        if metric == "cosine":
            distances = cosine_distances(X)
        if metric == "correlation":
            distances = 1 - np.corrcoef(X)
            vmin, vmax = -1, 1
        if metric == "euclidean":
            distances = euclidean_distances(X)
            vmin, vmax = 0, np.max(distances)

        fig, ax = plt.subplots()
        plt.imshow(distances, cmap=plt.cm.get_cmap().reversed(), vmin=vmin, vmax=vmax)
        plt.xticks(range(len(self)), self.embeddings.keys())
        plt.yticks(range(len(self)), self.embeddings.keys())
        plt.colorbar()

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")

    def plot_pixels(self):
        """
        Makes a pixelchart of every embedding in the set.

        Usage:

        ```python
        from whatlies.language import SpacyLanguage
        from whatlies.transformers import Pca

        lang = SpacyLanguage("en_core_web_sm")

        names = ['red', 'blue', 'green', 'yellow',
                 'cat', 'dog', 'mouse', 'rat',
                 'bike', 'car', 'motor', 'cycle',
                 'firehydrant', 'japan', 'germany', 'belgium']
        emb = lang[names].transform(Pca(12)).filter(lambda e: 'pca' not in e.name)
        emb.plot_pixels()
        ```

        ![](https://rasahq.github.io/whatlies/images/pixels.png)
        """
        names = self.embeddings.keys()
        df = self.to_dataframe()
        plt.matshow(df)
        plt.yticks(range(len(names)), names)

    def plot_movement(
        self,
        other,
        x_axis: Union[str, Embedding],
        y_axis: Union[str, Embedding],
        first_group_name="before",
        second_group_name="after",
        annot: bool = True,
    ):
        """
        Makes highly interactive plot of the movement of embeddings
        between two sets of embeddings.

        Arguments:
            other: the other embeddingset
            x_axis: the x-axis to be used, must be given when dim > 2
            y_axis: the y-axis to be used, must be given when dim > 2
            first_group_name: the name to give to the first set of embeddings (default: "before")
            second_group_name: the name to give to the second set of embeddings (default: "after")
            annot: drawn points should be annotated

        **Usage**

        ```python
        from whatlies.language import SpacyLanguage

        words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
                 "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
                 "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
                 "person", "family", "brother", "sister"]

        lang = SpacyLanguage("en_core_web_sm")
        emb = lang[words]
        emb_new = emb - emb['king']

        emb.plot_movement(emb_new, 'man', 'woman')
        ```
        """
        if isinstance(x_axis, str):
            x_axis = self[x_axis]
        if isinstance(y_axis, str):
            y_axis = self[y_axis]

        df1 = (
            self.to_axis_df(x_axis, y_axis).set_index("original").drop(columns=["name"])
        )
        df2 = (
            other.to_axis_df(x_axis, y_axis)
            .set_index("original")
            .drop(columns=["name"])
            .loc[lambda d: d.index.isin(df1.index)]
        )
        df_draw = (
            pd.concat([df1, df2])
            .reset_index()
            .sort_values(["original"])
            .assign(constant=1)
        )

        plots = []
        for idx, grp_df in df_draw.groupby("original"):
            _ = (
                alt.Chart(grp_df)
                .mark_line(color="gray", strokeDash=[2, 1])
                .encode(x="x_axis:Q", y="y_axis:Q")
            )
            plots.append(_)
        p0 = reduce(lambda x, y: x + y, plots)

        p1 = (
            deepcopy(self)
            .add_property("group", lambda d: first_group_name)
            .plot_interactive(x_axis, y_axis, annot=annot, color="group")
        )
        p2 = (
            deepcopy(other)
            .add_property("group", lambda d: second_group_name)
            .plot_interactive(x_axis, y_axis, annot=annot, color="group")
        )
        return p0 + p1 + p2

    def plot_interactive(
        self,
        x_axis: Union[int, str, Embedding] = 0,
        y_axis: Union[int, str, Embedding] = 1,
        axis_metric: Optional[Union[str, Callable, Sequence]] = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        title: Optional[str] = None,
        annot: bool = True,
        color: Union[None, str] = None,
    ):
        """
        Makes highly interactive plot of the set of embeddings.

        Arguments:
            x_axis: the x-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            y_axis: the y-axis to be used, must be given when dim > 2; if an integer, the corresponding
                dimension of embedding is used.
            axis_metric: the metric used to project each embedding on the axes; only used when the corresponding
                axis (i.e. `x_axis` or `y_axis`) is a string or an `Embedding` instance. It could be a string
                (`'cosine_similarity'`, `'cosine_distance'` or `'euclidean'`), or a callable that takes two vectors as input
                and returns a scalar value as output. To set different metrics for x- and y-axis, a list or a tuple of
                two elements could be given. By default (`None`), normalized scalar projection (i.e. `>` operator) is used.
            x_label: an optional label used for x-axis; if not given, it is set based on `x_axis` value.
            y_label: an optional label used for y-axis; if not given, it is set based on `y_axis` value.
            title: an optional title for the plot; if not given, it is set based on `x_axis` and `y_axis` values.
            annot: drawn points should be annotated
            color: a property that will be used for plotting

        **Usage**

        ```python
        from whatlies.language import SpacyLanguage

        words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
                 "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
                 "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
                 "person", "family", "brother", "sister"]

        lang = SpacyLanguage("en_core_web_sm")
        emb = lang[words]

        emb.plot_interactive('man', 'woman')
        ```
        """
        if isinstance(x_axis, str):
            x_axis = self[x_axis]
        if isinstance(y_axis, str):
            y_axis = self[y_axis]

        if isinstance(axis_metric, (list, tuple)):
            x_axis_metric = axis_metric[0]
            y_axis_metric = axis_metric[1]
        else:
            x_axis_metric = axis_metric
            y_axis_metric = axis_metric

        # Determine axes values and labels
        if isinstance(x_axis, int):
            x_val = self.to_X()[:, x_axis]
            x_lab = "Dimension " + str(x_axis)
        else:
            x_axis_metric = Embedding._get_plot_axis_metric_callable(x_axis_metric)
            x_val = self.compare_against(x_axis, mapping=x_axis_metric)
            x_lab = x_axis.name

        if isinstance(y_axis, int):
            y_val = self.to_X()[:, y_axis]
            y_lab = "Dimension " + str(y_axis)
        else:
            y_axis_metric = Embedding._get_plot_axis_metric_callable(y_axis_metric)
            y_val = self.compare_against(y_axis, mapping=y_axis_metric)
            y_lab = y_axis.name
        x_label = x_label if x_label is not None else x_lab
        y_label = y_label if y_label is not None else y_lab
        title = title if title is not None else f"{x_lab} vs. {y_lab}"

        plot_df = pd.DataFrame(
            {
                "x_axis": x_val,
                "y_axis": y_val,
                "name": [v.name for v in self.embeddings.values()],
                "original": [v.orig for v in self.embeddings.values()],
            }
        )

        if color:
            plot_df[color] = [
                getattr(v, color) if hasattr(v, color) else ""
                for v in self.embeddings.values()
            ]

        result = (
            alt.Chart(plot_df)
            .mark_circle(size=60)
            .encode(
                x=alt.X("x_axis", axis=alt.Axis(title=x_label)),
                y=alt.X("y_axis", axis=alt.Axis(title=y_label)),
                tooltip=["name", "original"],
                color=alt.Color(":N", legend=None) if not color else alt.Color(color),
            )
            .properties(title=title)
            .interactive()
        )

        if annot:
            text = (
                alt.Chart(plot_df)
                .mark_text(dx=-15, dy=3, color="black")
                .encode(
                    x="x_axis",
                    y="y_axis",
                    text="original",
                )
            )
            result = result + text
        return result

    def plot_interactive_matrix(
        self,
        *axes: Union[int, str, Embedding],
        axes_metric: Optional[Union[str, Callable, Sequence]] = None,
        annot: bool = True,
        width: int = 200,
        height: int = 200,
    ):
        """
        Makes highly interactive plot of the set of embeddings.

        Arguments:
            axes: the axes that we wish to plot; each could be either an integer, the name of
                an existing embedding, or an `Embedding` instance (default: `0, 1`).
            axes_metric: the metric used to project each embedding on the axes; only used when the corresponding
                axis is a string or an `Embedding` instance. It could be a string (`'cosine_similarity'`,
                `'cosine_distance'` or `'euclidean'`), or a callable that takes two vectors as input and
                returns a scalar value as output. To set different metrics for different axes, a list or a tuple of
                the same length as `axes` could be given. By default (`None`), normalized scalar projection (i.e. `>` operator) is used.
            annot: drawn points should be annotated
            width: width of the visual
            height: height of the visual

        **Usage**

        ```python
        from whatlies.language import SpacyLanguage
        from whatlies.transformers import Pca

        words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
                 "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
                 "dog", "cat", "mouse", "red", "bluee", "green", "yellow", "water",
                 "person", "family", "brother", "sister"]

        lang = SpacyLanguage("en_core_web_sm")
        emb = lang[words]

        emb.transform(Pca(3)).plot_interactive_matrix(0, 1, 2)
        ```
        """
        # Set default value of axes, if not given.
        if len(axes) == 0:
            axes = [0, 1]

        if isinstance(axes_metric, (list, tuple)) and len(axes_metric) != len(axes):
            raise ValueError(
                f"The number of given axes metrics should be the same as the number of given axes. Got {len(axes)} axes vs. {len(axes_metric)} metrics."
            )
        if not isinstance(axes_metric, (list, tuple)):
            axes_metric = [axes_metric] * len(axes)

        # Get values of each axis according to their type.
        axes_vals = {}
        X = self.to_X()
        for axis, metric in zip(axes, axes_metric):
            if isinstance(axis, int):
                vals = X[:, axis]
                axes_vals["Dimension " + str(axis)] = vals
            else:
                if isinstance(axis, str):
                    axis = self[axis]
                metric = Embedding._get_plot_axis_metric_callable(metric)
                vals = self.compare_against(axis, mapping=metric)
                axes_vals[axis.name] = vals

        plot_df = pd.DataFrame(axes_vals)
        plot_df["name"] = [v.name for v in self.embeddings.values()]
        plot_df["original"] = [v.orig for v in self.embeddings.values()]
        axes_names = list(axes_vals.keys())

        result = (
            alt.Chart(plot_df)
            .mark_circle()
            .encode(
                x=alt.X(alt.repeat("column"), type="quantitative"),
                y=alt.Y(alt.repeat("row"), type="quantitative"),
                tooltip=["name", "original"],
            )
        )
        if annot:
            text_stuff = result.mark_text(dx=-15, dy=3, color="black").encode(
                text="original",
            )
            result = result + text_stuff

        result = (
            result.properties(width=width, height=height)
            .repeat(row=axes_names[::-1], column=axes_names)
            .interactive()
        )

        return result



from whatlies.transformers._transformer import Transformer
from whatlies.transformers._pca import Pca
from whatlies.transformers._umap import Umap
from whatlies.transformers._noise import Noise
from whatlies.transformers._addrandom import AddRandom
from whatlies.transformers._tsne import Tsne
from whatlies.transformers._normalizer import Normalizer
from whatlies.error import NotInstalled

try:
    from whatlies.transformers._opentsne import OpenTsne
except ModuleNotFoundError:
    OpenTsne = NotInstalled("OpenTsne", "opentsne")

try:
    from whatlies.transformers._ivis import Ivis
except ModuleNotFoundError:
    Ivis = NotInstalled("Ivis", "ivis")

__all__ = [
    "Transformer",
    "Pca",
    "Umap",
    "Noise",
    "AddRandom",
    "Tsne",
    "OpenTsne",
    "Ivis",
    "Normalizer",
]
